from litellm.llms.openai.image_generation import DallE2ImageGenerationConfig


class AzureDallE2ImageGenerationConfig(DallE2ImageGenerationConfig):
    """
    Azure dall-e-2 image generation config
    """

    pass
